@@ -20,7 +20,7 @@
 
 #include "katestatusbar.h"
 
-#include "katemodemenu.h"
+#include "katemodemenulist.h"
 #include "kateglobal.h"
 #include "katemodemanager.h"
 #include "katedocument.h"
@@ -28,6 +28,7 @@
 #include "kateabstractinputmode.h"
 #include "wordcounter.h"
 
+#include <KActionCollection>
 #include <KLocalizedString>
 #include <KIconUtils>
 #include <KAcceleratorManager>
@@ -186,13 +187,21 @@ KateStatusBar::KateStatusBar(KTextEditor::ViewPrivate *view)
     m_encoding->setMenu(m_view->encodingAction()->menu());
     m_encoding->setWhatsThis(i18n("Encoding"));
 
+    /**
+     * load the mode menu, which contains a scrollable list + search bar.
+     * This is an alternative menu to the mode action menu of the view.
+     */
+    KateModeMenuList *modeMenuList = new KateModeMenuList(i18n("Mode"), this);
+    modeMenuList->setWhatsThis(i18n("Here you can choose which mode should be used for the current document. This will influence the highlighting and folding being used, for example."));
+    modeMenuList->updateMenu(m_view->doc());
     /**
      * add mode button which allows user to switch mode of document
      * this will reuse the mode action menu of the view
      */
     m_mode = new StatusBarButton(this);
     topLayout->addWidget(m_mode);
-    m_mode->setMenu(m_view->modeAction()->menu());
+    modeMenuList->setButton(m_mode, false, KateModeMenuList::Inverse);
+    m_mode->setMenu(modeMenuList);
     m_mode->setWhatsThis(i18n("Syntax highlighting"));
 
     // signals for the statusbar
@@ -222,6 +231,19 @@ void KateStatusBar::contextMenuEvent(QContextMenuEvent *event)
 {
     // TODO Add option "Show Statusbar" and options to show/hide buttons of the status bar
     QMenu menu(this);
+
+    if (childAt(event->pos()) == m_inputMode) {
+        if (QAction *inputModesAction = m_view->actionCollection()->action(QStringLiteral("view_input_modes"))) {
+            if (QMenu *inputModesMenu = inputModesAction->menu()) {
+                const auto actions = inputModesMenu->actions();
+                for (int i = 0; i < actions.count(); ++i) {
+                    menu.addAction(actions.at(i));
+                }
+                menu.addSeparator();
+            }
+        }
+    }
+
     QAction *showLines = menu.addAction(QStringLiteral("Show line count"), this, &KateStatusBar::toggleShowLines);
     showLines->setCheckable(true);
     showLines->setChecked(KateViewConfig::global()->showLineCount());
@@ -234,7 +256,7 @@ void KateStatusBar::contextMenuEvent(QContextMenuEvent *event)
 
 void KateStatusBar::toggleShowLines(bool checked)
 {
-    KateViewConfig::global()->setShowLineCount(checked);
+    KateViewConfig::global()->setValue(KateViewConfig::ShowLineCount, checked);
 }
 
 void KateStatusBar::toggleShowWords(bool checked)
